"""
ldapsession.py - higher-level class for handling LDAP connections
(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: ldapsession.py,v 1.200 2003/05/15 07:52:43 michael Exp $
"""

__version__ = '0.5.2'

import sys,time,types,ldap,ldap.cidict,ldaputil.base

from ldap.ldapobject import ReconnectLDAPObject as LDAPObject

# Constants for options
OPT_OFF=0
OPT_ON=1
OPT_DICT={'referrals':ldap.OPT_REFERRALS,'restart':ldap.OPT_RESTART}

OPTIONS_ATTRS = ['referrals','restart','deref']

START_TLS_NO = 0
START_TLS_TRY = 1
START_TLS_REQUIRED = 2

# Used attributes from RootDSE
ROOTDSE_ATTRS = [
  'namingContexts',
  'defaultNamingContext',
  'defaultRnrDN',
  'altServer',
  'ogSupportedProfile',
  'supportedSASLMechanisms',
  'supportedLDAPVersion',
  'supportedControl',
  'subschemaSubentry',
  'supportedFeatures',
  'vendorName','vendorVersion',
]

READ_CACHE_EXPIRE = 900

class LDAPSession:
  """
  Class for handling LDAP connection objects
  """

  def __init__(
    self,on_behalf=None,traceLevel=0,traceFile=None
  ):
    """Initialize a LDAPSession object"""
    self.namingContexts = []
    self._dn = u''
    self._traceLevel = traceLevel
    self._traceFile = traceFile or sys.stdout
    # Character set/encoding of data stored on this particular host
    self.charset = 'utf-8'
    # This is a dictionary for storing arbitrary objects
    # tied to a LDAP session
    self.read_cache = {}
    self.schema_dn_cache = {}
    self.schema_cache = {}
    # Default timeout 60 seconds
    self.timeout = 60
    # Initial state of Manage DSA IT mode
    self.manageDsaItenabled = 0
    # Capable of returning only attribute types with a search call
    self.onlyAttrTypes = 1
    # Supports feature described in draft-zeilenga-ldap-opattrs
    self.supportsAllOpAttr = 0
    # IP adress, host name or other free form information
    # of proxy client
    self.onBehalf = on_behalf

  def _retryConnect(
    self,uri,trace_level,trace_file,max_conn_try=1,conn_try_delay=0.5
  ):
    """
    Try connecting for max_conn_try times with pausing
    conn_try_delay seconds between failed connection attempts.
    """
    trial_counter = 0
    while 1:
      trial_counter += 1
      try:
        l = LDAPObject(uri,trace_level,trace_file)
      except ldap.LDAPError:
        if trial_counter>=max_conn_try:
          raise
        time.sleep(conn_try_delay)
      else:
        break
    return l

  def _supportedLDAPVersion(self):
    """
    Try to determine the highest supported protocol version
    by trying to bind anonymously
    """
    # Set protocol version to LDAPv3
    self.l.set_option(ldap.OPT_PROTOCOL_VERSION,ldap.VERSION3)
    # first try LDAPv3 bind
    try:
      # Try to bind to provoke error reponse if protocol
      # version is not supported
      self.l.bind_s('','',ldap.AUTH_SIMPLE)
    except ldap.PROTOCOL_ERROR,e:
      # Make sure that error just happened because of wrong
      # protocol version
      if hasattr(e,'args') and \
         type(e.args)==type(()) and \
         type(e.args[0])==type({}) and \
         e.args[0].get('info','').lower()=='version not supported':
        # Drop connection completely
        self.l.unbind_s() ; del self.l
        # Reconnect to host
        self.l = self._retryConnect(
          self.uri,self._traceLevel,self._traceFile
        )
        # Switch to new connection to LDAPv2
        self.l.set_option(ldap.OPT_PROTOCOL_VERSION,ldap.VERSION2)
        self.l.bind_s('','',ldap.AUTH_SIMPLE)
        self.who = ''
      else:
        # Raise any other error exception
        raise
      # Set currently determined protocol version
      version = 2
    else:
      self.who = ''
      version = 3
    return version

  def setTLSOptions(
    self,
    # Options for verifying SSL server certificate
    tls_cacertdir='',tls_cacertfile='',
    # Options for deploying client certificates for strong authentication
    tls_certfile='',tls_keyfile='',
  ):
    # Only set the options if python-ldap was built with TLS support
    if hasattr(ldap,'OPT_X_TLS_CACERTDIR') and tls_cacertdir and not tls_cacertfile:
      ldap.set_option(ldap.OPT_X_TLS_CACERTDIR,tls_cacertdir)
    if hasattr(ldap,'OPT_X_TLS_CACERTFILE') and tls_cacertfile and not tls_cacertdir:
      ldap.set_option(ldap.OPT_X_TLS_CACERTFILE,tls_cacertfile)
    if hasattr(ldap,'OPT_X_TLS_CERTFILE') and tls_certfile:
      ldap.set_option(ldap.OPT_X_TLS_CERTFILE,tls_certfile)
    if hasattr(ldap,'OPT_X_TLS_KEYFILE') and tls_keyfile:
      ldap.set_option(ldap.OPT_X_TLS_KEYFILE,tls_keyfile)

  def startTLSExtOp(self,startTLSOption=START_TLS_TRY):
    """StartTLS if possible and requested"""
    if startTLSOption:
      self.l.set_option(ldap.OPT_PROTOCOL_VERSION,ldap.VERSION3)
      self.l.set_option(ldap.OPT_X_TLS,startTLSOption)
      try:
        self.l.start_tls_s()
      except (ldap.SERVER_DOWN,ldap.CONNECT_ERROR):
        self.secureConn = 0
        raise
      except AttributeError,e:
        self.secureConn = 0
        raise ldap.SERVER_DOWN(
          {
            'desc':str(e),
            'info':'python-ldap installation is lacking StartTLS support'
          }
        )
      else:
        self.secureConn = 1
    else:
      self.secureConn = 0

  def _initialize(self,uri_list):
    while uri_list:
      uri = uri_list[0].strip().encode('ascii')
      # Try connecting to LDAP host
      try:
        self.l = self._retryConnect(
          uri,self._traceLevel,self._traceFile
        )
      except ldap.LDAPError:
        # Remove current host from list
        uri_list.pop(0)
        if uri_list:
          # Try next host
          continue
        else:
          raise
      else:
        break
    return uri

  def open(
    self,uri,timeout=60,
    startTLS=START_TLS_TRY,
    # Options for verifying SSL server certificate
    tls_cacertdir='',tls_cacertfile='',
    # Options for deploying client certificates for strong authentication
    tls_certfile='',tls_keyfile='',
  ):
    """
    Open a LDAP connection with separate DNS lookup
    
    uri
        Either a (Unicode) string or a list of strings
        containing LDAP URLs of host(s) to connect to.
        If host is a list connecting is tried until a
        connect to a host in the list was successful.
    """

    assert uri,ValueError("No host string or list specified for %s.open()." % (
        self.__class__.__name__
      ))
    assert type(uri) in [types.StringType,types.UnicodeType,types.ListType], \
      TypeError("Parameter uri must be either list of strings or single string.")
    self.timeout = timeout
    if type(uri) in [types.StringType,types.UnicodeType]:
      uri_list = [uri]
    elif type(uri)==types.ListType:
      uri_list = uri
    self.setTLSOptions(tls_cacertdir,tls_cacertfile,tls_certfile,tls_keyfile)
    self.uri = self._initialize(uri_list)
    if self.uri.lower().startswith('ldap:'):
      # Start TLS extended operation
      self.startTLSExtOp(startTLS)
    elif self.uri.lower().startswith('ldaps:') or self.uri.lower().startswith('ldapi:'):
      self.secureConn = 1
    # Get and save currently used LDAP protocol version
    self._supportedLDAPVersion()


  def unbind(self):
    """Close LDAP connection object if necessary"""
    if hasattr(self,'l'):
      try:
        self.l.unbind_s()
      except ldap.LDAPError:
        pass
      except AttributeError:
        pass
      del self.l
    try:
      del self.uri # delete the LDAP connection URI
    except AttributeError:
      pass    

  def getUmichConfig(self):
    """
    Try to read entry cn=config attribute database
    from UMich LDAPv2 server derivate
    """
    try:
      ldap_result = self.readEntry(
        'cn=config',['database'],
      )
    except ldap.NO_SUCH_OBJECT:
      result = []
    except ldap.PARTIAL_RESULTS:
      result = []
    except ldap.UNDEFINED_TYPE:
      result = []
    except ldap.INAPPROPRIATE_MATCHING:
      result = []
    except ldap.OPERATIONS_ERROR:
      result = []
    else:
      if ldap_result:
	result = []
	l = ldap_result[0][1].get('database',[])
	for d in l:
          try:
            dbtype,basedn = d.split(' : ',1)
	  except ValueError:
	    pass
	  else:
            result.extend(
              [
                dn.strip()
                for dn in basedn.split(' : ')
              ]
            )
      else:
	result = []
    return [
      unicode(i,self.charset)
      for i in result
    ]

  def _forgetRootDSEAttrs(self):
    """Forget all old RootDSE values"""
    for attrtype in ROOTDSE_ATTRS:
      if hasattr(self,attrtype):
        delattr(self,attrtype)
    self.supportsAllOpAttr = 0

  def getRootDSE(self):
    """Retrieve attributes from Root DSE"""
    self._forgetRootDSEAttrs()
    self.namingContexts = []
    try:
      ldap_result = self.readEntry(
        '',ROOTDSE_ATTRS+['objectClass']
      )
    except ldap.NO_SUCH_OBJECT:
      pass
    except ldap.PARTIAL_RESULTS:
      pass
    except ldap.UNDEFINED_TYPE:
      pass
    except ldap.INAPPROPRIATE_MATCHING:
      pass
    except ldap.OPERATIONS_ERROR:
      pass
    except ldap.UNWILLING_TO_PERFORM:
      pass
    else:
      if ldap_result:
        rootDSE=ldap.cidict.cidict(ldap_result[0][1])
      else:
        rootDSE=ldap.cidict.cidict()
      for attrtype in ROOTDSE_ATTRS:
        attrvalue = ldaputil.base.unicode_list(
          rootDSE.get(attrtype,rootDSE.get(attrtype.lower(),[])),
          self.charset
        )
        if attrvalue:
          if attrtype=='vendorName' or attrtype=='vendorVersion':
            # These are single-valued attributes by definition
            setattr(self,attrtype,attrvalue[0])
          else:
            setattr(self,attrtype,attrvalue)
      self.supportsAllOpAttr = \
        ('1.3.6.1.4.1.4203.1.5.1' in rootDSE.get('supportedFeatures',[])) or \
        ('OpenLDAProotDSE' in rootDSE.get('objectClass',[]))
      schema_dn_list = rootDSE.get('subschemaSubEntry',[])
      # Speed up sub schema sub entry retrieval by pre-filling cache
      # with what is likely the sub schema for whole DIT
      if schema_dn_list and schema_dn_list[0]!=None:
        self.schema_dn_cache[u''] = unicode(schema_dn_list[0],self.charset)
        self.retrieveSubSchema(u'',None)
      else:
        self.schema_dn_cache[u''] = None
    if not hasattr(self,'namingContexts') or not self.namingContexts:
      self.namingContexts=self.getUmichConfig()
    self.namingContexts = map(ldaputil.base.normalize_dn,self.namingContexts[:])
    return # getRootDSE()

  def getSearchRoot(self,dn):
    """
    Returns the namingContexts value matching best the
    distinguished name given in dn
    """
    return ldaputil.base.match_dnlist(dn,self.namingContexts)

  def isLeafEntry(self,dn):
    """Returns 1 if the node is a leaf entry, 0 otherwise"""
    # First try to read operational attributes from entry itself
    # which might indicate whether there are subordinate entries
    result_ldap = self.l.search_st(
      dn.encode(self.charset),
      ldap.SCOPE_BASE,'(objectClass=*)',
      ['hasSubordinates','subordinateCount'],
      timeout=self.timeout
    )
    if result_ldap:
      entry = ldap.cidict.cidict(result_ldap[0][1])
      if entry.has_key('hasSubordinates'):
        return entry['hasSubordinates'][0].upper()=='FALSE'
      elif entry.has_key('subordinateCount'):
        return int(entry['subordinateCount'][0])==0
    ldap_msgid = self.l.search(
      dn.encode(self.charset),
      ldap.SCOPE_ONELEVEL,'(objectClass=*)',
      ['objectClass'],self.onlyAttrTypes
    )
    result_ldap = (None,None)
    while result_ldap==(None,None):
      result_ldap = self.l.result(ldap_msgid,0,self.timeout)
    self.l.abandon(ldap_msgid)
    return not result_ldap[1]

  def getObjectClasses(self,dn):
    """Get a list of object classes associated with an entry"""
    search_result = self.readEntry(dn,['objectClass'])
    if search_result:
      entry = search_result[0][1]
      return entry.get('objectClass',entry.get('objectclass',[]))
    else:
      raise ldap.NO_SUCH_OBJECT
    return [] # getObjectClasses()

  def getSubschemaEntryDN(self,dn):
    """Get DN of subschemaSubentry for dn"""
    # Search for DN of subschemaSubentry
    return self.l.search_subschemasubentry_s(dn.encode(self.charset))

  def searchSubSchemaEntryDN(self,dn):
    """Determine DN of sub schema sub entry for current part of DIT"""
    if self.schema_dn_cache.has_key(dn):
      # grab DN of sub schema sub entry from schema DN cache
      subschemasubentry_dn = self.schema_dn_cache[dn]
    else:
      # Search the DN of sub schema sub entry
      subschemasubentry_dn = self.l.search_subschemasubentry_s(dn.encode('utf-8'))
      # Store DN of sub schema sub entry in schema DN cache
      self.schema_dn_cache[dn] = subschemasubentry_dn
    return subschemasubentry_dn

  def retrieveSubSchema(self,dn,default):
    """Retrieve parsed sub schema sub entry for current part of DIT"""
    if self.l.protocol_version<ldap.VERSION3:
      # LDAPv2 gets back default schema
      return default
    else:
      subschemasubentry_dn = self.searchSubSchemaEntryDN(dn)
      if subschemasubentry_dn is None:
        # No sub schema sub entry found => return default schema
        return default
      elif self.schema_cache.has_key(subschemasubentry_dn):
        # Return parsed schema from cache
        sub_schema = self.schema_cache[subschemasubentry_dn]
      else:
        # Read the sub schema sub entry
        subschemasubentry = self.l.read_subschemasubentry_s(
          subschemasubentry_dn.encode(self.charset),ldap.schema.SCHEMA_ATTRS
        )
        if subschemasubentry is None:
          sub_schema = None
        else:
          # Parse the schema
          sub_schema = ldap.schema.SubSchema(subschemasubentry)
        # Store parsed schema in schema cache
        self.schema_cache[subschemasubentry_dn] = sub_schema
      # Determine what to return
      return sub_schema or default

  def getSubschemaEntry(self,dn):
    """Read the whole subschemaSubentry for dn"""
    subschemaSubentryDN = self.getSubschemaEntryDN(dn)
    if subschemaSubentryDN:
      # Read the whole schema entry
      search_result = self.readEntry(
        subschemaSubentryDN,
        ['objectClasses','attributeTypes','matchingRules','ldapSyntaxes'],
        '(objectClass=subschema)'
      )
      if search_result:
        return search_result[0]
      else:
        raise ldap.NO_SUCH_OBJECT
    else:
      raise ldap.NO_SUCH_OBJECT

  def getAttributeTypes(self,dn,attrs=None,brute_force=0,opAttrs=1):
    """Get a list of attributes present in an entry"""
    attrs = attrs or []
    if not dn:
      request_attrs = ldap.cidict.strlist_union(attrs,ROOTDSE_ATTRS)
    else:
      request_attrs = attrs
    # Do the wildcard search first
    search_result = self.readEntry(
      dn,{0:None,1:['*','+']}[opAttrs and self.supportsAllOpAttr],
      self.onlyAttrTypes
    )
    if search_result:
      result = search_result[0][1].keys()
    else:
      result = []
    if request_attrs:
      try:
        # Explicitly search for certain attributes
        search_result = self.readEntry(
          dn,request_attrs,self.onlyAttrTypes
        )
      except ldap.NO_SUCH_ATTRIBUTE:
        if brute_force:
          for attr in request_attrs:
            try:
              # Explicitly search attribute-wise which is pretty hefty!
              search_result = self.readEntry(
                dn,[attr],self.onlyAttrTypes
              )
            except ldap.NO_SUCH_ATTRIBUTE:
              pass
            else:
              result.extend(search_result[0][1].keys())
      else:
        if search_result:
          result = ldap.cidict.strlist_union(result,search_result[0][1].keys())
    return result # getAttributeTypes()

  def readEntry(
    self,dn,attrtype_list=None,only_attrtypes=0,
    search_filter='(objectClass=*)',no_cache=0
  ):
    """Read a single entry"""
    if attrtype_list==['*']:
      attrtype_list = None
    acid = ','.join(attrtype_list or ['__'])
    if not no_cache and \
       self.read_cache.has_key(dn) and \
       self.read_cache[dn].has_key(acid):
      timestamp,read_cache_result = self.read_cache[dn][acid]
      if timestamp+READ_CACHE_EXPIRE>time.time():
        return read_cache_result
      else:
        del self.read_cache[dn][acid]
    # Read single entry from LDAP server
    search_result = self.l.search_st(
      dn.encode(self.charset),ldap.SCOPE_BASE,search_filter,
      attrtype_list,only_attrtypes,self.timeout
    )
    if search_result:
      # Create DN-level cache dictionary
      if not self.read_cache.has_key(dn):
        self.read_cache[dn]={}
      # Store the read entry in the time-stamped read_cache
      self.read_cache[dn][acid] = (time.time(),search_result)
    return search_result

  def existingEntry(self,dn,suppress_referrals=0):
    """Returns 1 if entry exists, 0 if NO_SUCH_OBJECT was raised."""
    try:
      self.readEntry(dn,[])
    except ldap.NO_SUCH_OBJECT:
      return 0
    except ldap.PARTIAL_RESULTS:
      if suppress_referrals:
        return 0
      else:
        raise
    else:
      return 1
      
  def flushCache(self):
    """Flushes all LDAP cache data"""
    self.read_cache = {}
    self.schema_dn_cache = {}
    self.schema_cache = {}

  def uncacheEntry(self,dn):
    """Removes all cached items of entry from cache"""
    try:
      del self.read_cache[dn]
    except KeyError:
      pass

  def addEntry(self,dn,modlist):
    """Add single entry"""
    self.l.add_s(dn.encode(self.charset),modlist)
    return

  def modifyEntry(self,dn,modlist):
    """Modify single entry"""
    if modlist:
      self.uncacheEntry(dn)
      self.l.modify_s(dn.encode(self.charset),modlist)
    return # modifyEntry()

  def renameEntry(self,dn,new_rdn,new_superior=None,delold=1):
    """Rename an entry"""
    self.uncacheEntry(dn)
    old_superior_str = ldaputil.base.ParentDN(ldaputil.base.normalize_dn(dn))
    if new_superior!=None:
      if old_superior_str==ldaputil.base.normalize_dn(new_superior):
        new_superior_str = None
      else:
        new_superior_str = new_superior.encode(self.charset)
    self.l.rename_s(
      dn.encode(self.charset),new_rdn.encode(self.charset),
      new_superior_str,delold
    )
    return u','.join([new_rdn,new_superior or old_superior]) # renameEntry()

  def deleteEntry(self,dn):
    """Delete single entry"""
    self.uncacheEntry(dn)
    self.l.delete_s(dn.encode(self.charset))
    return # deleteEntry()

  def setDN(self,dn):
    """
    Set currently used DN.
    """
    if type(dn)==types.StringType:
      dn = unicode(dn,self.charset)
    dn=ldaputil.base.normalize_dn(dn)
    if dn!=self._dn:
      self._dn = dn

  def manageDsaIt(self,enable):
    """
    Enable or disable manageDSAit mode (see draft-zeilenga-ldap-namedref)
    Mainly this is a wrapper which saves the state.
    
    If enable is None nothing is done
    """
    if self.l.protocol_version>=ldap.VERSION3:
      self.l.manage_dsa_it(enable)
      self.manageDsaItenabled = enable

  def bind(
    self,
    who='',
    cred='',
    method=ldap.AUTH_SIMPLE,
    filtertemplate='(uid=%s)',
    loginSearchRoot=''
  ):
    """
    Send BindRequest to LDAP server
    """
    # Smart login
    if filtertemplate and who and not ldaputil.base.is_dn(who):
      who = unicode(ldaputil.base.SmartLogin(
        self.l,
        who.encode(self.charset),
        loginSearchRoot,
        filtertemplate,
        attrnamesonly=self.onlyAttrTypes,
        timeout=self.timeout
      ),self.charset)

    # Call the bind
    try:
      self.l.bind_s(who.encode(self.charset),cred.encode(self.charset),method)
    finally:
      # Make sure reference to credential is deleted in any case
      del cred
    # Store bind DN
    self.who = who
    # Flush schema cache
    self.schema_dn_cache = {}
    self.schema_cache = {}
    # Flush old data from cache
    self.flushCache()
    # Access to root DSE might have changed after binding
    # as another entity
    self.getRootDSE()
    return

  def valid(self):
    """return 1 if connection is valid"""
    return hasattr(self,'l')

  def __repr__(self):
    if hasattr(self,'l'):
      connection_str = (' LDAPv%d' % (self.l.protocol_version))
    else:
      connection_str = ''
    return '<LDAPSession%s:%s>' % (
      connection_str,
      ','.join(
        [
          '%s:%s' % (a,repr(getattr(self,a)))
          for a in ['uri','who','dn','onBehalf','startedTLS']
          if hasattr(self,a)
        ]
      )
    )
