"""
w2lapp.handler.py: base handler

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: handler.py,v 1.201 2003/05/30 17:40:42 michael Exp $
"""

import sys,os,types,re,socket,\
       ldap,ldaputil.base,ldapdns,ldapsession,mssignals, \
       pyweblib.forms,pyweblib.httphelper,pyweblib.sslenv,pyweblib.helper,pyweblib.session

# Import the application modules
import w2lapp.core,w2lapp.gui,w2lapp.cnf, \
       w2lapp.passwd,w2lapp.searchform,w2lapp.locate, \
       w2lapp.search,w2lapp.addmodifyform,w2lapp.add, \
       w2lapp.modify,w2lapp.delete, \
       w2lapp.read,w2lapp.conninfo,w2lapp.login,w2lapp.connect, \
       w2lapp.referral,w2lapp.monitor,w2lapp.groupadm,w2lapp.rename, \
       w2lapp.srvrr,w2lapp.schema.viewer

from ldapurl import LDAPUrl,isLDAPUrl
from ldapsession import LDAPSession
from w2lapp.core import utf2display

from w2lapp.form import Web2LDAPForm,FORM_CLASS

try:
  SocketErrors = (socket.error,socket.gaierror)
except AttributeError:
  SocketErrors = socket.error

def CheckRequiredSecLevel(ls,env):
  required_ssl_level = w2lapp.cnf.GetParam(ls,'ssl_minlevel',0)
  current_ssl_level = pyweblib.sslenv.SecLevel(
    env,
    w2lapp.cnf.misc.sec_sslacceptedciphers,
    w2lapp.cnf.GetParam(ls,'ssl_valid_dn',''),
    w2lapp.cnf.GetParam(ls,'ssl_valid_idn','')
  )
  if current_ssl_level < required_ssl_level:
    raise w2lapp.core.ErrorExit(
      'Access denied. SSL security level %d not sufficient. Must be at least %d.' % (
        current_ssl_level,required_ssl_level
      )
    )


def CommandDispatcher(sid,outf,command,form,connLDAPUrl,ls,dn,env):
  """Execute function for command"""
  if command=='searchform':
    w2lapp.searchform.w2l_SearchForm(sid,outf,command,form,ls,dn)
  elif command=='search':
    w2lapp.search.w2l_Search(
      sid,outf,command,form,ls,dn,
      connLDAPUrl.scope,
      w2lapp.core.str2unicode(connLDAPUrl.filterstr,form.accept_charset),
    )
  elif command=='addform' or command=='modifyform':
    w2lapp.addmodifyform.w2l_AddModifyForm(sid,outf,command,form,ls,dn)
  elif command=='add':
    w2lapp.add.w2l_Add(sid,outf,command,form,ls,dn)
  elif command=='modify':
    w2lapp.modify.w2l_Modify(sid,outf,command,form,ls,dn)
  elif command=='delete':
    w2lapp.delete.w2l_Delete(sid,outf,command,form,ls,dn,connLDAPUrl.scope)
  elif command=='rename':
    w2lapp.rename.w2l_Rename(sid,outf,command,form,ls,dn)
  elif command=='passwd':
    w2lapp.passwd.w2l_Passwd(sid,outf,command,form,ls,dn)
  elif command=='read':
    w2lapp.read.w2l_Read(
      sid,outf,command,form,ls,dn,
      wanted_attrs={
        0:connLDAPUrl.attrs,1:[]
      }[connLDAPUrl.attrs is None],
    )
  elif command=='conninfo':
    w2lapp.conninfo.w2l_ConnInfo(sid,outf,command,form,ls,dn,env)
  elif command=='login':
    w2lapp.login.w2l_Login(
      sid,outf,'searchform',form,ls,dn,connLDAPUrl.scope,
      form.getInputValue('login_search_root',[ls.getSearchRoot(dn)])[0]
    )
  elif command=='groupadm':
    w2lapp.groupadm.w2l_GroupAdm(sid,outf,command,form,ls,dn)
  elif command=='oid':
    w2lapp.schema.viewer.w2l_DisplaySchemaElement(sid,outf,command,form,ls,dn)
  return # CommandDispatcher()


def ExtractPathInfo(env):
  # Extract the command from PATH_INFO env var
  path_info = env.get('PATH_INFO','/')[1:]
  if not path_info:
    c,s = '',''
  else:
    # Work around broken web servers which adds the script name
    # to path info as well
    script_name = env['SCRIPT_NAME']
    if path_info.startswith(script_name):
      path_info = path_info[len(script_name):]
    try:
      c,s = path_info.split('/',1)
    except ValueError:
      c,s = path_info,''
    if c=='ldapurl':
      c,s = '',''
  return c,s # extractPathInfo()


def HandleHTTPRequest(inf,outfile,errfile,env,script_name=None):

  ls = None

  ####################################################
  # Main try-except block for catching and logging
  # all unhandled exceptions
  ####################################################

  try:

    command,sid = ExtractPathInfo(env)

    form = FORM_CLASS.get(command,Web2LDAPForm)(inf,env,script_name or env['SCRIPT_NAME'])

    outf = form.outFileObject(outfile)
    
    if not FORM_CLASS.has_key(command):
      pyweblib.httphelper.URLRedirect(outf,form.script_name,2,'Invalid web2ldap command %s!' % repr(command))
      return

    dn = None

    ####################################################
    # try-except block for gracefully handling of
    # certain exceptions (mainly w2lapp.core.ErrorExit and ldap.LDAPError)
    ####################################################

    try:

      if command=='urlredirect':
        # URL redirecting has absolutely nothing to do with rest
        pyweblib.httphelper.URLRedirect(outfile,env.get('QUERY_STRING',''))
        return
      elif command=='monitor':
        # Output simple monitor page. Does not require session handling.
        w2lapp.monitor.w2l_Monitor(outf,command,form)
        return
      elif command=='locate':
        form.getInputFields(ignoreEmptyFields=0)
        w2lapp.locate.w2l_Locate(outf,command,form)
        return
      elif command=='disconnect':
        # Explicitly remove old session
        w2lapp.core.session.deleteSession(sid)
        # Redirect to start page to avoid people bookmarking disconnect URL
        pyweblib.httphelper.URLRedirect(outf,form.script_name,0,'Disconnect...')
        return
      elif command=='':
        # New connect => create new session, remove old if necessary
        w2lapp.core.session.deleteSession(sid)
        # Just output a connect form if there was not query string
        if not form.query_string:
          w2lapp.connect.w2l_Connect(outf,command,form,LDAPSession())
          return

      ####################################################
      # Restore old or initialize new web session object
      ####################################################

      try:

        if sid:

          #################################################
          # Session ID given => try to restore old session
          #################################################

          try:
            ls = w2lapp.core.session.retrieveSession(sid,env)
          except pyweblib.session.SessionExpired,e:
            # Handle expired sessions
            form.getInputFields(ignoreEmptyFields=0)
            if hasattr(e.session_data,'uri'):
              # Allow to relogin
              sid = w2lapp.core.session.newSession(env)
              ls = e.session_data ; ls.cred = ''
              dn = form.getInputValue('dn',[ls._dn])[0]
              w2lapp.login.w2l_Login(
                sid,outf,command,form,ls,dn,None,
                form.getInputValue('login_search_root',[ls.getSearchRoot(dn)])[0],
                login_msg=str(e),
                who=ls.__dict__.get('who',''),
                relogin=1,nomenu=1
              )
            else:
              pyweblib.httphelper.URLRedirect(outf,form.script_name,2,'Session invalid and expired.')
            return
          except pyweblib.session.KeyError:
            pyweblib.httphelper.URLRedirect(outf,form.script_name,2,'Session ID not found.')
            return
          except pyweblib.session.SessionHijacked:
            pyweblib.httphelper.URLRedirect(
              outf,form.script_name,2,'Session hijacking detected. Access denied!'
            )
            return
          else:
            if ls=='_created_':
              # In case of dummy session retrieved we have to create
              # LDAPSession instance
              ls = LDAPSession(
                on_behalf=w2lapp.core.guessClientAddr(env),
                traceLevel=w2lapp.cnf.misc.ldap_trace_level,traceFile=errfile
              )

        else:

          #################################################
          # No session ID given => create new session
          #################################################

          sid = w2lapp.core.session.newSession(env)
          ls = LDAPSession(
            on_behalf=w2lapp.core.guessClientAddr(env),
            traceLevel=w2lapp.cnf.misc.ldap_trace_level,traceFile=errfile
          )

      except pyweblib.session.MaxSessionCountExceeded:
        pyweblib.httphelper.SimpleMsg(
          outf,'Too many web sessions! Try later...'
        )
        return

      if isLDAPUrl(form.query_string):
        # Extract the connection parameters from a LDAP URL
        try:
          inputLDAPUrl = LDAPUrl(form.query_string)
        except ValueError,e:
          raise w2lapp.core.ErrorExit('Error parsing LDAP URL: %s.' % (str(e)))

      else:

        # Extract the connection parameters from form fields
        form.getInputFields(ignoreEmptyFields=0)

        if 'ldapurl' in form.inputFieldNames:
          # One form parameter with LDAP URL
          ldap_url_input = form.field['ldapurl'].value[0]
          try:
            inputLDAPUrl = LDAPUrl(ldap_url_input.encode('ascii'))
          except ValueError,e:
            raise w2lapp.core.ErrorExit('Error parsing LDAP URL: %s.' % (str(e)))
        else:
          inputLDAPUrl = LDAPUrl(hostport=form.getInputValue('host',[None])[0])

      # Separate parameters for dn, who, cred and scope
      # have predecence over parameters specified in LDAP URL
      inputLDAPUrl.dn=form.getInputValue('dn',[inputLDAPUrl.dn])[0]
      inputLDAPUrl.who=form.getInputValue('who',[inputLDAPUrl.who])[0]
      inputLDAPUrl.cred=form.getInputValue('cred',[inputLDAPUrl.cred])[0]

      dn = w2lapp.core.str2unicode(inputLDAPUrl.dn,form.accept_charset)
      who = w2lapp.core.str2unicode(inputLDAPUrl.who,form.accept_charset)
      cred = w2lapp.core.str2unicode(inputLDAPUrl.cred,form.accept_charset)
      
      scope_str=form.getInputValue(
        'scope',
        [
          {0:str(inputLDAPUrl.scope),1:''}[inputLDAPUrl.scope is None]
        ]
       )[0]
      if scope_str:
        inputLDAPUrl.scope=int(scope_str)
      else:
        inputLDAPUrl.scope=ldap.SCOPE_BASE

      inputLDAPUrl.scope = inputLDAPUrl.scope or ldap.SCOPE_BASE

      if not command:
        command = {0:'search',1:'read'}[inputLDAPUrl.scope==ldap.SCOPE_BASE]

      ####################################################
      # Connect to LDAP host
      ####################################################

      if inputLDAPUrl.hostport:

        if not inputLDAPUrl.hostport and \
           inputLDAPUrl.urlscheme=='ldap' and \
           not hasattr(ls,'uri'):
          # Force a SRV RR lookup for dc-style DNs,
          # create list of URLs to connect to
          initializeUrl = [
            LDAPUrl(urlscheme='ldap',hostport='%s' % (host)).initializeUrl()
            for host in ldapdns.dcDNSLookup(dn)
          ]
          if not initializeUrl:
            # No hosts found via DNS
            pyweblib.httphelper.URLRedirect(
              outf,form.script_name,2,'No host specified.'
            )
            return
        else:
          initializeUrl = str(inputLDAPUrl.initializeUrl()[:])


        if (not hasattr(ls,'uri') or initializeUrl!=ls.uri):
          # Connect to new specified host
          del ls
          ls = LDAPSession(
            on_behalf=w2lapp.core.guessClientAddr(env),
            traceLevel=w2lapp.cnf.misc.ldap_trace_level,traceFile=errfile
          )
          try:
            ls.open(
              initializeUrl,
              startTLS=w2lapp.cnf.GetParam(inputLDAPUrl,'starttls',ldapsession.START_TLS_NO),
              tls_cacertdir=w2lapp.cnf.GetParam(inputLDAPUrl,'tls_cacertdir',''),
              tls_cacertfile=w2lapp.cnf.GetParam(inputLDAPUrl,'tls_cacertfile',''),
              tls_certfile=w2lapp.cnf.GetParam(inputLDAPUrl,'tls_certfile',''),
              tls_keyfile=w2lapp.cnf.GetParam(inputLDAPUrl,'tls_keyfile',''),
            )
          except ldap.SERVER_DOWN,e:
            # Connecting impossible => remove session
            w2lapp.core.session.deleteSession(sid)
            # Redirect to entry page
            pyweblib.httphelper.URLRedirect(
              outf,form.script_name,2,
              'Connecting to %s impossible: %s' % (initializeUrl,str(e))
            )
            return
          ls.l.set_option(ldap.OPT_RESTART,0)
          ls.l.set_option(ldap.OPT_DEREF,0)
          ls.l.set_option(ldap.OPT_REFERRALS,0)
          ls.timeout = w2lapp.cnf.GetParam(ls,'timeout',-1)
          ls.l.timeout = ls.timeout
          ls.getRootDSE()
          ls.setDN(dn)
          # Search and set backend specific timeout
          ls.timeout = w2lapp.cnf.GetParam(ls,'timeout',-1)
          ls.l.timeout = ls.timeout

      # Store session data in case anything goes wrong after here
      # to give the exception handler a good chance
      w2lapp.core.session.storeSession(sid,ls)

      ####################################################
      # Bind to LDAP host
      ####################################################

      if who!=None:

        loginSearchRoot = form.getInputValue(
          'login_search_root',[ls.getSearchRoot(dn)]
        )[0]

        if cred!=None:

          try:
            ls.bind(
              {0:'',1:who}[len(cred)>0],cred,
              filtertemplate=form.getInputValue('login_filterstr',[''])[0],
              loginSearchRoot = loginSearchRoot
            )
          except ldap.NO_SUCH_OBJECT,e:
            w2lapp.login.w2l_Login(
              sid,outf,command,form,ls,dn,inputLDAPUrl.scope,loginSearchRoot,
              login_msg=w2lapp.gui.LDAPError2ErrMsg(e),
              who=who,relogin=1
            )
            return
          else:
            who = ls.who.encode(ls.charset)

        else:
          w2lapp.login.w2l_Login(
            sid,outf,command,form,ls,dn,inputLDAPUrl.scope,loginSearchRoot,
            login_msg='',
            who=who,relogin=1,nomenu=1
          )
          return

      # Check for valid LDAPSession and connection to provide reasonable
      # error message instead of logging exception in case user is playing
      # with manually generated URLs
      if not isinstance(ls,ldapsession.LDAPSession) or not hasattr(ls,'uri'):
        pyweblib.httphelper.URLRedirect(outf,form.script_name,2,'No valid LDAP connection!')
        return

      ls.setDN(dn)

      # Store session data in case anything goes wrong after here
      # to give the exception handler a good chance
      w2lapp.core.session.storeSession(sid,ls)

      # Check backend specific required SSL level
      CheckRequiredSecLevel(ls,env)

      # Execute the command module
      try:
        CommandDispatcher(sid,outf,command,form,inputLDAPUrl,ls,dn,env)
      except ldap.SERVER_DOWN:
        ls.l.reconnect(ls.uri)
        CommandDispatcher(sid,outf,command,form,inputLDAPUrl,ls,dn,env)

    except pyweblib.forms.FormException,e:
      if ls is None:
        dn = None
      else:
        dn = ls.__dict__.get('_dn',None)
      w2lapp.gui.ExceptionMsg(
        sid,outf,form,ls,dn,'Error parsing form','Error parsing form: %s' % (e.html())
      )

    except ldap.NO_SUCH_OBJECT,e:

      #########################################
      # Generic handler for "No such object"
      #########################################

      host_list = ldapdns.dcDNSLookup(dn)
      if (not host_list) or (LDAPUrl(ls.uri).hostport in host_list):
        # Did not find another LDAP server for this naming context
        try:
          if type(e.args[0])==types.DictType and e.args[0].has_key('matched'):
            new_dn = unicode(e.args[0]['matched'],ls.charset)
          else:
            new_dn = dn
        except IndexError:
          new_dn = dn
        w2lapp.gui.ExceptionMsg(
          sid,outf,form,ls,new_dn,'No such object',
          '<br>'.join([
            w2lapp.gui.LDAPError2ErrMsg(e),
            w2lapp.gui.DisplayDN(sid,form,ls,dn),
          ])
        )
      else:
        # Found LDAP server for this naming context via DNS SRV RR
        w2lapp.srvrr.w2l_ChaseSRVRecord(sid,outf,command,form,ls,dn,host_list)

    except (ldap.PARTIAL_RESULTS,ldap.REFERRAL),e:
      w2lapp.referral.w2l_ChaseReferral(sid,outf,command,form,ls,dn,e)

    except (ldap.INSUFFICIENT_ACCESS,ldap.STRONG_AUTH_REQUIRED),e:
      w2lapp.login.w2l_Login(
        sid,outf,command,form,ls,dn,inputLDAPUrl.scope,
        form.getInputValue('login_search_root',[ls.getSearchRoot(dn)])[0],
        who='',
        login_msg=w2lapp.gui.LDAPError2ErrMsg(e),relogin=1
      )

    except (ldap.INVALID_CREDENTIALS,ldap.INAPPROPRIATE_AUTH),e:
      who = who or ls.__dict__.get('who','')
      w2lapp.login.w2l_Login(
        sid,outf,command,form,ls,dn,inputLDAPUrl.scope,
        form.getInputValue('login_search_root',[ls.getSearchRoot(dn)])[0],
        login_msg=w2lapp.gui.LDAPError2ErrMsg(e),
        who=who,relogin=1
      )

    except ldap.SERVER_DOWN,e:
      # Try to re-establish an old connection
      uri=ls.uri
      who=ls.__dict__.get('who','')
      ls.unbind(); del ls
      ls = LDAPSession(
        on_behalf=w2lapp.core.guessClientAddr(env),
        traceLevel=w2lapp.cnf.misc.ldap_trace_level,traceFile=errfile
      )
      try:
        # Try to reconnect
        ls.uri = uri # Needed for reading starttls from config
        ls.open(
          uri,
          startTLS=ls.__dict__.get(
            'startedTLS',
            w2lapp.cnf.GetParam(ls,'starttls',ldapsession.START_TLS_NO)
          )
        )
      except ldap.LDAPError,e:
        # Reconnecting impossible => remove session
        w2lapp.core.session.deleteSession(sid)
        # Redirect to entry page
        pyweblib.httphelper.URLRedirect(
          outf,form.script_name,2,
          'Reconnecting to %s impossible.' % (uri)
        )
        return
      else:
        # Reconnecting was possible => new connect
        ls.l.set_option(ldap.OPT_DEREF,0);
        ls.l.set_option(ldap.OPT_REFERRALS,0)
        ls.setDN(dn)
        w2lapp.login.w2l_Login(
          sid,outf,command,form,ls,dn,inputLDAPUrl.scope,
          form.getInputValue('login_search_root',[ls.getSearchRoot(dn)])[0],
          login_msg='Connection %s invalid. Retry with new login.' % (ls.uri.encode('ascii')),
          who=who,relogin=1,nomenu=1
        )
    except ldaputil.base.USERNAME_NOT_FOUND,e:
      w2lapp.login.w2l_Login(
        sid,outf,command,form,ls,dn,inputLDAPUrl.scope,
        form.getInputValue('login_search_root',[ls.getSearchRoot(dn)])[0],
        login_msg=w2lapp.gui.LDAPError2ErrMsg(e),
        who=who,relogin=1
      )
    except ldaputil.base.USERNAME_NOT_UNIQUE,e:
      login_search_root = form.getInputValue('login_search_root',[ls.getSearchRoot(dn)])[0]
      w2lapp.login.w2l_Login(
        sid,outf,command,form,ls,dn,inputLDAPUrl.scope,
        login_search_root,
        login_msg=' | '.join([
          w2lapp.gui.LDAPError2ErrMsg(e),
          form.applAnchor(
	    'search','Show',sid,
            [
              ('dn',login_search_root),
              ('scope',str(ldap.SCOPE_SUBTREE)),
              ('search_filterstr',w2lapp.cnf.GetParam(ls,'binddnsearch',r'(uid=%s)').replace('%s',who))
            ]
        )]),
        who=who,relogin=1
      )

    except ldap.TIMEOUT,e:
      if __debug__:
        w2lapp.core.log_exception(errfile,ls,env)
      w2lapp.gui.ExceptionMsg(
        sid,outf,form,ls,dn,'LDAP timeout',
        'Timeout of %d secs exceeded: %s' % (ls.l.timeout,w2lapp.gui.LDAPError2ErrMsg(e))
      )

    except ldap.LDAPError,e:
      if __debug__:
        w2lapp.core.log_exception(errfile,ls,env)
      ErrMsg = w2lapp.gui.LDAPError2ErrMsg(e)
      w2lapp.gui.ExceptionMsg(sid,outf,form,ls,dn,'LDAP exception',ErrMsg)

    except mssignals.SigPipeException,e:
      # SIGPIPE received
      if __debug__:
        w2lapp.core.log_exception(errfile,ls,env)

    except SocketErrors,e:
      w2lapp.gui.ExceptionMsg(sid,outf,form,None,None,'Socket Error',str(e))

    except UnicodeError,e:
      if __debug__:
        w2lapp.core.log_exception(errfile,ls,env)
      else:
        w2lapp.gui.ExceptionMsg(sid,outf,form,None,None,'Unicode Error',str(e))

    except IOError,e:
      if e.errno in [24,32,104]:
        raise
      else:
        w2lapp.gui.ExceptionMsg(sid,outf,form,ls,dn,'I/O Error',str(e))

    except w2lapp.core.ErrorExit,e:
      w2lapp.gui.ExceptionMsg(sid,outf,form,ls,dn,'Error',e.Msg)

    w2lapp.core.session.storeSession(sid,ls)

  except IOError,e:
    if e.errno in [24,32,104]:
      raise

  except:
    # Log unhandled exceptions
    w2lapp.core.log_exception(errfile,ls,env)

  return # HandleHTTPRequest()
