"""
w2lapp.passwd.py: change password associated with entry

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: passwd.py,v 1.32 2003/04/12 18:58:12 michael Exp $
"""

import time,ldap,ldaputil.base,ldaputil.passwd, \
       w2lapp.cnf,w2lapp.core,w2lapp.gui,w2lapp.login

available_hashtypes = ldaputil.passwd.AVAIL_USERPASSWORD_SCHEMES.items()


try:
  import md4
except ImportError:
  md4 = None
else:
  import binascii

def ntPasswordHash(password):
  """
  Returns MD4-based NT password hash for a password.
  """
  md4_digest = md4.new()
  md4_digest.update(password[:128].encode('utf-16-le'))
  return binascii.hexlify(md4_digest.digest()).upper()


def PasswdForm(
  sid,outf,command,form,ls,dn,passwd_who,user_objectclasses,
  Heading,Msg
):

  config_hashtypes = w2lapp.cnf.GetParam(ls,'passwd_hashtypes',[])
  if config_hashtypes:
    # The set of hash types are restricted by local configuration
    default_hashtypes = [
      hash_type
      for hash_type in available_hashtypes
      if hash_type[0] in config_hashtypes
    ]
    form.field['passwd_scheme'].options = default_hashtypes

  if 'sambaaccount' in user_objectclasses or 'smbPasswordEntry' in user_objectclasses:
    if md4:
      ntpassword_sync_field = form.field['passwd_ntpasswordsync'].inputHTML()
    else:
      ntpassword_sync_field = 'not available in local installation'
    ntpassword_sync_tablerow = """<tr>
      <td>Sync with ntPassword attribute (for Samba)</td>
      <td>%s</td>
    </tr>
    """ % (ntpassword_sync_field)
  else:
    ntpassword_sync_tablerow = ''

  w2lapp.gui.TopSection(
    sid,outf,form,ls,dn,'Change password',
    w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
  )
  outf.write("""<div id="Input" class="Main">
<h1>%s</h1>
<p>%s</p>
%s
<input type="hidden" type="dn" name="dn" value="%s">
<input type="hidden" name="passwd_who" value="%s">
<table summary="Password input form">
<tr>
  <td>Change password for</td>
  <td>%s on %s</td>
</tr>
<tr>
  <td>New password</td>
  <td><input type="password" name="passwd_newpasswd" maxlength="30" size="30" value=""></td>
</tr>
<tr>
  <td>repeat new password</td>
  <td><input type="password" name="passwd_newpasswd" maxlength="30" size="30" value=""></td>
</tr>
<tr>
  <td>Password encryption method</td>
  <td>%s</td>
</tr>
%s
</table>
<input type="submit" value="Change Password">
<input type="reset" value="Reset">
</form>
</div>
""" % (
  Heading,Msg,
  form.beginFormHTML('passwd',sid,'POST'),
  w2lapp.core.utf2display(form.accept_charset,dn),
  w2lapp.core.utf2display(form.accept_charset,passwd_who),
  w2lapp.gui.DisplayDN(sid,form,ls,passwd_who),
  ls.uri.encode('ascii'),
  form.field['passwd_scheme'].inputHTML(),
  ntpassword_sync_tablerow
)
)
  w2lapp.gui.PrintFooter(outf,form)
  return # PasswdForm()


def w2l_Passwd(sid,outf,command,form,ls,dn):

  passwd_who = form.getInputValue('passwd_who',[ls.who or dn])[0]

  user_objectclasses = [ oc.lower() for oc in ls.getObjectClasses(passwd_who) ]

  if passwd_who and \
     ('passwd_newpasswd' in form.inputFieldNames):

    ####################################################################
    # New password provided => (re)set it in entry
    ####################################################################

    sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))
    required_attrs_dict,allowed_attrs_dict = sub_schema.attribute_types(
      user_objectclasses,strict=0,raise_keyerror=0
    )
    if not required_attrs_dict.has_key('userPassword') and \
       not allowed_attrs_dict.has_key('userPassword') and \
       not 'simplesecurityobject' in user_objectclasses and \
       sub_schema.name2oid[ldap.schema.ObjectClass].has_key('simpleSecurityObject'):
      modlist = [(ldap.MOD_ADD,'objectClass','simpleSecurityObject')]
    else:
      modlist = []

    if len(form.field['passwd_newpasswd'].value)!=2:
      raise w2lapp.core.ErrorExit('Repeat password!')

    if form.field['passwd_newpasswd'].value[0]!=form.field['passwd_newpasswd'].value[1]:
      PasswdForm(
        sid,outf,command,form,ls,dn,passwd_who,user_objectclasses,
        Heading='Password Error',Msg='New passwords do not match!'
      )
      return

    passwd_input = form.field['passwd_newpasswd'].value[0]

    if not passwd_input:
      PasswdForm(
        sid,outf,command,form,ls,dn,passwd_who,user_objectclasses,
        Heading='Password Error',Msg='Empty passwords not allowed!'
      )
      return

    if 'passwd_scheme' in form.inputFieldNames:
      passwd_scheme = form.field['passwd_scheme'].value[0]
    else:
      raise w2lapp.core.ErrorExit('Password encryption method required!')

    userPassword = ldaputil.passwd.UserPassword(
      l=ls.l,dn=passwd_who.encode(ls.charset),charset=ls.charset,multiple=0
    )
    # Work around the problem that LDAP applications use different
    # charset encodings for the userPassword attribute
    new_passwd_value = []
    for c in ['utf-8']:#,'iso-8859-1']:
      try:
        new_passwd_str = passwd_input.encode(c)
      except UnicodeError:
        pass
      else:
        new_passwd_value.append(
          userPassword.encodePassword(
            new_passwd_str,passwd_scheme
          )
        )

    modlist.append((ldap.MOD_REPLACE,'userPassword',new_passwd_value))

    pwd_change_timestamp = time.time()

    if 'shadowaccount' in user_objectclasses:
      modlist.append((ldap.MOD_REPLACE,'shadowLastChange',str(int(pwd_change_timestamp/86400))))

    if form.getInputValue('passwd_ntpasswordsync',['no'])[0]=='yes':
      modlist.append((ldap.MOD_REPLACE,'ntPassword',ntPasswordHash(passwd_input)))
      modlist.append((ldap.MOD_REPLACE,'pwdLastSet',str(int(pwd_change_timestamp))))

    # Modify password
    ls.modifyEntry(passwd_who,modlist)

    # Check if relogin is necessary
    if ldaputil.base.normalize_dn(passwd_who.lower())==ldaputil.base.normalize_dn(ls.who.lower()):
      ls.bind('','')
      w2lapp.login.w2l_Login(
        sid,outf,'searchform',form,ls,dn,None,ls.getSearchRoot(passwd_who),
        login_msg='Password changed of entry %s. A new login is required!' % (w2lapp.gui.DisplayDN(sid,form,ls,passwd_who)),
        who=passwd_who,relogin=0,nomenu=1
      )
    else:
      w2lapp.gui.SimpleMessage(
        sid,outf,form,ls,dn,
        message='Changed password of entry %s.' % (
          w2lapp.gui.DisplayDN(sid,form,ls,passwd_who)
        ),
        main_menu_list=w2lapp.gui.MainMenu(sid,form,ls,dn),
        context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
      )

  else:

    ####################################################################
    # New password not yet provided => ask for it
    ####################################################################

    PasswdForm(sid,outf,command,form,ls,dn,passwd_who,user_objectclasses,'New password','Provide a new password.')

